/*
 * Copyright (c) 1984,1985,1989,1994,1995  Mark Nudelman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice in the documentation and/or other materials provided with 
 *    the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN 
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*
 * Routines dealing with getting input from the keyboard (i.e. from the user).
 */

#include "less.h"

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif /* WIN32 */

#ifdef WIN32
public HANDLE hConsoleIn;
#else /* not WIN32 */
static int tty;
#endif /* not WIN32 */

/*
 * Open keyboard for input.
 */
	public void
open_getchr()
{
#if MSOFTC || OS2
	extern int fd0;
	/*
	 * Open a new handle to CON: in binary mode 
	 * for unbuffered keyboard read.
	 */
	 fd0 = dup(0);
	 close(0);
	 tty = OPEN_TTYIN();
#else
#ifdef WIN32
	hConsoleIn = CreateFile ("CONIN$", GENERIC_READ | GENERIC_WRITE,
				 0, NULL, OPEN_EXISTING, 0, NULL);
	raw_mode (1);
#else /* not WIN32 */
	/*
	 * Try /dev/tty.
	 * If that doesn't work, use file descriptor 2,
	 * which in Unix is usually attached to the screen,
	 * but also usually lets you read from the keyboard.
	 */
	tty = OPEN_TTYIN();
	if (tty < 0)
		tty = 2;
#endif /* not WIN32 */
#endif
}

#ifdef WIN32

	public void
close_getchr (void)
{
  raw_mode (0);
  CloseHandle (hConsoleIn);
}

#endif /* WIN32 */


#ifdef WIN32

	static int
decode_key (KEY_EVENT_RECORD *event, WORD *code)
{
  DWORD dwControl = event->dwControlKeyState;
  BYTE keystate[256];
  int isdead;

  if (!event->bKeyDown)
    return -1;

  memset (keystate, 0, sizeof (keystate));
  if (dwControl & SHIFT_PRESSED)
    keystate[VK_SHIFT] = 0x80;
  if (dwControl & CAPSLOCK_ON)
    keystate[VK_CAPITAL] = 1;
  if (dwControl & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED))
    keystate[VK_CONTROL] = 0x80;

  isdead = ToAscii (event->wVirtualKeyCode, event->wVirtualScanCode,
		    keystate, code, 0);

  if (isdead == 0)
    {
      switch (event->wVirtualKeyCode)
	{
	case VK_LEFT: case VK_RIGHT: case VK_UP: case VK_DOWN:
	case VK_PRIOR: case VK_NEXT:
	case VK_HOME: case VK_END:
	case VK_INSERT: case VK_DELETE:
	case VK_F1: case VK_F2: case VK_F3: case VK_F4:
	case VK_F5: case VK_F6: case VK_F7: case VK_F8:
	case VK_F9: case VK_F10: case VK_F11: case VK_F12:
	  if (dwControl & SHIFT_PRESSED)
	    {
	      isdead = 3;
	      code[0] = (WORD) '\340';
	      code[1] = VK_SHIFT;
	      code[2] = event->wVirtualKeyCode;
	    }
	  else if (dwControl & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED))
	    {
	      isdead = 3;
	      code[0] = (WORD) '\340';
	      code[1] = VK_CONTROL;
	      code[2] = event->wVirtualKeyCode;
	    }
	  else
	    {
	      isdead = 2;
	      code[0] = (WORD) '\340';
	      code[1] = event->wVirtualKeyCode;
	    }
	  break;
	default:
	  isdead = -1;
	}
    }

  return isdead;
}

#endif /* WIN32 */

/*
 * Get a character from the keyboard.
 */
	public int
getchr()
{
	char c;
	int result;
#ifdef WIN32
	INPUT_RECORD event;
	DWORD dwRead;
	static WORD code[3];
	static int isdead;
	extern int sigs;
#endif /* WIN32 */

	do
	{
#if MSOFTC
		/*
		 * In raw read, we don't see ^C so look here for it.
		 */
		flush();
		c = getch();
		result = 1;
		if (c == '\003')
			return (READ_INTR);
#else
#if OS2
		flush();
		while (_read_kbd(0, 0, 0) != -1)
			continue;
		if ((c = _read_kbd(0, 1, 0)) == -1)
			return (READ_INTR);
		result = 1;
#else
#ifdef WIN32
		flush ();
		if (isdead == 3)
		  {
		    c = (char) code[1];
		    code[1] = code[2];
		    isdead = 2;
		    break;
		  }
		if (isdead == 2)
		  {
		    c = (char) code[1];
		    isdead = 0;
		    break;
		  }
		result = 0;
		WaitForSingleObject (hConsoleIn, INFINITE);
		if (sigs & S_INTERRUPT)
		  return READ_INTR;
		ReadConsoleInput (hConsoleIn, &event, 1, &dwRead);
		if (event.EventType == KEY_EVENT)
		  {
		    isdead = decode_key (&event.Event.KeyEvent, code);
		    if (isdead < 0)
		      continue;

		    c = (char) code[0];
		    result = 1;
		  }
#else /* not WIN32 */
		result = iread(tty, &c, sizeof(char));
		if (result == READ_INTR)
			return (READ_INTR);
		if (result < 0)
		{
			/*
			 * Don't call error() here,
			 * because error calls getchr!
			 */
			quit(QUIT_ERROR);
		}
#endif /* not WIN32 */
#endif
#endif
		/*
		 * Various parts of the program cannot handle
		 * an input character of '\0'.
		 * If a '\0' was actually typed, convert it to '\340' here.
		 */
		if (c == '\0')
			c = '\340';
	} while (result != 1);

	return (c & 0377);
}
